from abc import ABCMeta, abstractmethod

from sc2replaylib import Sc2replaylibException

class Parser:
	__metaclass__ = ABCMeta
	
	@abstractmethod
	def parse(self):
		pass

class ParserException(Sc2replaylibException):
	pass


from mpyq import MPQArchive

from sc2replaylib import Sc2replaylibException
from sc2replaylib.parsers.attributes import AttributesParser
from sc2replaylib.parsers.details import DetailsParser

class Replay:

	GAME_TEAMS = {
		'1v1': '1v1',
		'2v2': '2v2',
		'3v3': '3v3',
		'4v4': '4v4',
		'6v6': '6v6',
		'FFA': 'Free for all',
		'Cust': 'Custom Game'
	}
	
	GAME_SPEED = {
		'Slow': 'Slow',
		'Slor': 'Slower',
		'Norm': 'Normal',
		'Fasr': 'Faster',
		'Fast': 'Fast'
	}
	
	GAME_MATCHING = {
		'Priv':	'Private',
		'Amm':	'Auto Match Maker',
		'Pub':	'Public'
	}	
	
	FILES = {
		'attributes':	'replay.attributes.events',
		'details':		'replay.details'
	}

	def __init__(self, replay_file):
		
		self.teams 				= []
		self.replay_file		= replay_file
		
		self.details_parser		= None
		self.attributes_parser 	= None

		try:
			archive = MPQArchive(self.replay_file)
			files = archive.extract()

			# bootstrap the right parsers, expand here for different version parsing too
			for file_name, data in files.iteritems():
				if(file_name == self.FILES['attributes']):
					self.attributes_parser = AttributesParser(data)
						
				if(file_name == self.FILES['details']):
					self.details_parser = DetailsParser(data)
		
			teams = self.attribute(2001)
			num_teams = 2

			if teams == '1v1':
				teams_lookup_attribute = 2002
				
			elif teams == '2v2':
				teams_lookup_attribute = 2003
				
			elif teams == '3v3':
				teams_lookup_attribute = 2004
				
			elif teams == '4v4':
				teams_lookup_attribute = 2005
				
			elif teams == 'FFA':
				teams_lookup_attribute = 2006
				num_teams = 10
				
			elif teams == '6v6':
				teams_lookup_attribute = 2008
			
			# create the teams before the players
			for i in range(num_teams):
				self.teams.append(Team(i+1))
			
			# bootstrap the player object with some raw data
			for i, player_details in enumerate(self.details_parser.parse()[0]):
				
				player = Player(player_details, self.player_attributes(i+1))
								
				# team
				players_team = player.attribute(teams_lookup_attribute)								
				self.teams[int(players_team[1])-1].players.append(player)
				
		except IOError as (errno, strerror):
			print strerror
		
	def attribute(self, key):
		attributes = self.attributes()
		for attr in attributes:
			if attr[1] == key:
				return attr[3]

		# no attribute found!
		raise Sc2replaylibException("no global attribute found with key '%d'" % (key))
	
	def player_attributes(self, player_num):
		rc = []
		attributes = self.attributes_parser.parse()
		for attrib in attributes:
			if attrib[2] == player_num:
				rc.append(attrib)
		return rc
	
	def attributes(self):
		return self.player_attributes(16)

	@property
	def game_teams(self):
		return self.attribute(2001)
		
		
	@property
	def game_speed(self):
		return self.attribute(3000)
	
	@property
	def game_matching(self):
		return self.attribute(3009)
	
	@property
	def map_human_friendly(self):
		return self.details_parser.parse()[1]

	@property
	def timestamp(self):
		from datetime import datetime
		return datetime.fromtimestamp((self.details_parser.parse()[5] - 116444735995904000) / 10**7)

	@property
	def timezone_offset(self):
		return (self.details_parser.parse()[6] / 10**7 ) / (60 * 60)

class Team:
	
	OUTCOME	= {
		0: "Unknown",
		1: "Won",
		2: "Lossed"
	}
		
	def __init__(self, team_number):
		self.players = []
		self.number = team_number
		
	@property		
	def outcome(self):
		outcome = self.OUTCOME[0]
		for p in self.players:
			if outcome == self.OUTCOME[0] and p.outcome != self.OUTCOME[0]:
				outcome = p.outcome
			elif p.outcome != outcome:
				return self.OUTCOME[0]
		return outcome
	
class Player:
	
	RACE = {
		'Terran':		'Terran',
		'Protoss':		'Protoss',
		'Zerg':			'Zerg',
		'Rand-Terran':	'Random Terran',
		'Rand-Protoss':	'Random Protoss',
		'Rand-Zerg':	'Random Zerg'
	}
	
	OUTCOME	= {
		0: "Unknown",
		1: "Won",
		2: "Lossed"
	}
	
	COLORS = {
		'tc01': 'Red',
		'tc02': 'Blue',
		'tc03': 'Teal',
		'tc04': 'Purple',
		'tc05': 'Yellow',
		'tc06': 'Orange',
		'tc07': 'Green',
		'tc08': 'Light Pink',
		'tc09': 'Violet',
		'tc10': 'Light Grey',
		'tc11': 'Dark Green',
		'tc12': 'Brown',
		'tc13': 'Light Green',
		'tc14': 'Dark Grey',
		'tc15': 'Pink'
	}
	
	TYPE = {
		'Humn': 'Human',
		'Comp': 'Computer'
	}
	
	DIFFICULTY = {
		'Easy': 'Easy',
		'VyEy': 'Very Easy',
		'Medi': 'Medium',
		'VyHd': 'Very Hard',
		'Insa': 'Insane'
	}
	
	def __init__(self, details, attributes):
		self.details	= details
		self.attributes	= attributes

	@property
	def type(self):
		return self.attribute(500)

	@property
	def handle(self):
		return self.details[0]
		
	@property
	def bnet_id(self):
		return self.details[1]
		
	@property
	def race(self):
		race = self.attribute(3001)
		if race == 'RAND':
			if self.details[2] == 'Terran':
				return 'Rand-Terran'
			
			elif self.details[2] == 'Protoss':
				return 'Rand-Protoss'
			
			elif self.details[2] == 'Zerg':
				return 'Rand-Zerg'
		
		elif race == 'Terr':
			return 'Terran'
			
		elif race == 'Prot':
			return 'Protoss'
			
		elif rate == 'Zerg':
			return 'Zerg'
		
	@property
	def color_argb(self):
		return self.details[3]
		
	@property
	def color_name(self):
		return self.COLORS[self.attribute(3002)]
		
	@property
	def handicap(self):
		return	self.details[6]
	
	@property	
	def outcome(self):
		return self.details[8]
	
	def attribute(self, key):
		for attr in self.attributes:
			if attr[1] == key:
				return attr[3]
		
		# no attribute found!
		raise Sc2replaylibException("no attribute found for player '%d' with key '%d'" % (attr[2], key))


import struct

from sc2replaylib.parsers import Parser, ParserException

class AttributesParser(Parser):
				
		ATTRIB_BYTE_SIZE		= 13
		ATTRIB_HEADER_OFFSET	= 9
		
		def __init__(self, raw_data):
			self.raw_data = raw_data
			self.parsed_data = None
			
		def parse(self):
			if self.parsed_data == None:
			 	self.parsed_data = self.parse_events(self.raw_data)
			
			return self.parsed_data
		
		
		def unpack_int32(self, data):
			return struct.unpack_from('<I', data)[0]
		
		def unpack_str(self, data):
			return struct.unpack_from('4s',data)[0][::-1].replace('\x00','')
		
		# parses an attribute packet
		def parse_attribute(self, data):
			return tuple([
				self.unpack_int32(data[0:4]),
				self.unpack_int32(data[4:8]),
				ord(data[8:9]),
				self.unpack_str(data[9:13])])
		
		def parse_events(self, data):

			# attributes header, 9 bytes long
			# should look like: 00  00  00  00  00  aa  aa  aa  aa
			#                   ^ 5 bytes of zeros  ^ 4 byte int of total attribute length			
			self.length = self.unpack_int32(data[5:9])
			
			# parse 13 byte blocks until self.length is met
			i = 0
			parsed = []
			while i < self.length:
				parsed.append(self.parse_attribute(data[
					self.ATTRIB_HEADER_OFFSET+(self.ATTRIB_BYTE_SIZE*i):
					self.ATTRIB_HEADER_OFFSET+(self.ATTRIB_BYTE_SIZE*(i+1))]))
				i += 1
			
			return tuple(parsed)


from sc2replaylib.parsers import Parser, ParserException

class DetailsParser(Parser):

	DETAILS_TYPE_BIN			= '\x02'
	DETAILS_TYPE_ARRAY			= '\x04'
	DETAILS_TYPE_INDEXED_ARRAY	= '\x05'
	DETAILS_TYPE_TINYINT		= '\x06'
	DETAILS_TYPE_BIGINT			= '\x07'
	DETAILS_TYPE_VLF			= '\x09'
		
	def __init__(self, raw_data):
		self.raw_data		= raw_data
		self.parsed_data	= None
	
	def parse(self):
		if self.parsed_data == None:
			(self.parsed_data, length) = self.parse_details(self.raw_data)
			
		return self.parsed_data
	
	# Recursive function to parse_details
	def parse_details(self, data):
		i = 0
		
		# \x02 binary data
		if(data[i] == self.DETAILS_TYPE_BIN):		
			i += 1
			
			(length, movement) = self.parse_number(data[i:])
			i += movement + 1
						
			return (data[i : i + length], i + length)
			
		# \x04 array
		elif(data[i] == self.DETAILS_TYPE_ARRAY):
		
			# \x01 should follow identifyer...
			i += 1
			if data[i] != '\x01':
				raise ParserException('This should have been \x01 in array')

			# ... then \x00					
			i += 1
			if data[i] != '\x00':
				raise ParserException('This should have been \x00 in array')
				
			# ... then length
			i += 1
			(length, movement) = self.parse_number(data[i:])
			i += movement + 1
			
			# move up i to the data!
			j = 0
			return_array = []
			
			while j < length:
								
				(new_data, movement) = self.parse_details(data[i:])				
				i += movement
				
				return_array.append(new_data)
				
				j += 1
							
			return (return_array, i)
					
		# \x05 Indexed array
		elif(data[i] == self.DETAILS_TYPE_INDEXED_ARRAY):
			i += 1
			
			(length, movement) = self.parse_number(data[i:])
			i += movement + 1
			
			return_array = []
			nulls = 0
			curr_idx = 0
			
			while (curr_idx - nulls) < length:

				# get index and check for 'null' entries - THEY DO NOT COUNT AGAINST ARRAY LENGTH
				(idx, movement) = self.parse_number(data[i:])
				i += movement + 1
				
				while idx > (curr_idx):
					return_array.append(None)
					curr_idx += 1
					nulls += 1
					
				curr_idx = idx
				(new_data, movement) = self.parse_details(data[i:])				
				i += movement
				
				return_array.append(new_data)				
				curr_idx += 1

			return (return_array, i)
		
		elif(data[i] == self.DETAILS_TYPE_TINYINT):
			i += 1			
			return (ord(data[i]), i + 1)
		
		#\x07 Big Integer
		elif(data[i] == self.DETAILS_TYPE_BIGINT):
			i += 1
			
			hex_str = hex(ord(data[i])) + str(hex(ord(data[i+1]))[2:]) + str(hex(ord(data[i+2]))[2:]) + str(hex(ord(data[i+3]))[2:])
			
			return (int(hex_str, 16), i+4)
		
		# \x09 VLF number found
		elif(data[i] == self.DETAILS_TYPE_VLF):
			i += 1
						
			(value, movement) = self.parse_number(data[i:])
			return (value, i + movement+1)
			
	def parse_number(self, data):
		i = 0

		# is the next byte going to be influential on this number?
		while ord(data[i]) & 0x80 == 0x80:
			i += 1

		value = ord(data[i]) & 0x7F
		
		j = i-1
		while j >= 0:
			bit = ord(data[j]) & 0x7F
			value = value << 7
			value = value | bit
			j -= 1
		
		if value & 1 == 1:
			value = (value >> 1) * -1
		else:
			value = (value >> 1)
		
		return (value, i)



import setuptools

setuptools.setup(
	name="sc2replay-lib",
	version="0.1.2",
	license="MPL 1.1",
	author="Dominic Baranski",
	author_email="dominic.baranski@gmail.com",
	url="https://github.com/dcbarans/sc2replay-lib",
	description="Basic library for reading details about a Blizzard (TM) StarCraft(TM) 2 replay file",
	long_description=''.join(open("README.txt").readlines()),
	keywords=["starcraft", "sc", "starcraft 2","sc2","parser","replay"],
	classifiers=[
			"Environment :: Console",
			"Development Status :: 4 - Beta",
			"Programming Language :: Python",
			"Programming Language :: Python :: 2.7",
			"Intended Audience :: Developers",
			"License :: OSI Approved :: Mozilla Public License 1.1 (MPL 1.1)",
			"Natural Language :: English",
			"Operating System :: OS Independent",
			"Environment :: Other Environment",
			"Topic :: Utilities",
			"Topic :: Software Development :: Libraries",
			"Topic :: Games/Entertainment :: Real Time Strategy",
		],
	setup_requires=['argparse'],
	install_requires=['mpyq>=0.1.8', 'argparse'],
	packages=['sc2replaylib', 'sc2replaylib.parsers'],
)
